@@ -4,9 +4,15 @@ extern "C" {
 
 // https://refspecs.linuxbase.org/LSB_3.1.0/LSB-Desktop-generic/LSB-Desktop-generic/libjpegman.html
 
+#include <math.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+
+#ifdef USE_TURBO
+#include <jmorecfg.h>
+#endif
+
 #include <jpeglib.h>
 
 #define DO_FANCY_UPSAMPLING 0x1
@@ -24,7 +30,6 @@ extern "C" {
 #define CCIR601_SAMPLING 0x1000
 
 GLOBAL(long) jround_up (long a, long b);
-//#include "cstegojpeg.h"
 
 FILE *_read_jpeg(const char *filename,
                  struct jpeg_decompress_struct *cinfo,
@@ -84,8 +89,9 @@ int read_jpeg_info(
     //num_components[1] = cinfo.out_color_components;
     //num_components[2] = cinfo.output_components;
   }
+  //fprintf(stderr, "setting jpeg color space %d %d\n", cinfo.jpeg_color_space, cinfo.out_color_space);
   if(jpeg_color_space != NULL)
-    jpeg_color_space[0] = cinfo.jpeg_color_space;
+    jpeg_color_space[0] = cinfo.out_color_space;
 
   if(samp_factor != NULL)
     for(int comp = 0; comp < cinfo.num_components; comp++) {
@@ -120,7 +126,6 @@ int read_jpeg_dct(
 
   // read DCT
   jvirt_barray_ptr *coeffs_array = jpeg_read_coefficients(&cinfo);
-
   // read dct
   JBLOCKARRAY buffer_one;
   JCOEFPTR blockptr_one;
@@ -144,13 +149,19 @@ int read_jpeg_dct(
       }
     }
   }
-  
+  //fprintf(stderr, "Read qt.\n");
   // read quantization table
   if(qt != NULL) {
     for(int ch = 0; ch < 2; ch++) {
+      //fprintf(stderr, "qt[%d] = %p -> %p\n", ch, cinfo.quant_tbl_ptrs[ch], cinfo.quant_tbl_ptrs[ch]->quantval);
+      for(int i = 0; i < 64; i++) {
+        
+        qt[ch*64 + i] = cinfo.quant_tbl_ptrs[ch]->quantval[i];//[(i&7)*8+(i>>3)];
+      }
+      //(i&7)*8+(i>>3)
       //memcpy((void *)(qt + ch*64), (void *)cinfo.quant_tbl_ptrs[ch]->quantval, sizeof(short)*64);
-      JQUANT_TBL *tbl = cinfo.comp_info[ch].quant_table;
-      memcpy((void *)(qt + ch*64), (void*)tbl->quantval, sizeof(short)*64);
+      //JQUANT_TBL *tbl = cinfo.comp_info[ch].quant_table;
+      //memcpy((void *)(qt + ch*64), (void*)tbl->quantval, sizeof(short)*64);
     }
   }
 
@@ -197,82 +208,154 @@ int write_jpeg_dct(
     fprintf(stderr, "you must specify either srcfile, qt or quality\n");
     return 0;
   }
+  if((srcfile == NULL) && (dct == NULL)) {
+    fprintf(stderr, "you must specify either srcfile or dct\n");
+    return 0;
+  }
 
   // allocate
   struct jpeg_compress_struct cinfo_out;
   struct jpeg_error_mgr jerr_out;
-  memset((void *)&cinfo_out, 0x0, sizeof(struct jpeg_compress_struct));
-  memset((void *)&jerr_out, 0x0, sizeof(struct jpeg_error_mgr));
-  
+  //memset((void *)&cinfo_out, 0x0, sizeof(struct jpeg_compress_struct));
+  //memset((void *)&jerr_out, 0x0, sizeof(struct jpeg_error_mgr));
+
   // open the destination file
   FILE * fp_out;
   if ((fp_out = fopen(dstfile, "wb")) == NULL) {
     fprintf(stderr, "can't open %s\n", dstfile);
     return 0;
   }
-  
-  cinfo_out.err = jpeg_std_error(&jerr_out);
-  jpeg_create_compress(&cinfo_out);
-  jpeg_stdio_dest(&cinfo_out, fp_out);
-  
-  // set basic parameters
-  cinfo_out.image_width = image_dims[0];
-  cinfo_out.image_height = image_dims[1];
-  cinfo_out.in_color_space = in_color_space;
-  if(in_components >= 0) cinfo_out.input_components = in_components;
-  cinfo_out.num_components = cinfo_out.input_components;
-  jpeg_set_defaults(&cinfo_out);
 
   // allocate
   struct jpeg_decompress_struct cinfo_in;
   struct jpeg_error_mgr jerr_in;
   FILE * fp_in;
   // read source jpeg
-  if(srcfile != NULL) {
+  if(srcfile != NULL)
     if((fp_in = _read_jpeg(srcfile, &cinfo_in, &jerr_in)) == NULL) return 0;
-    // copy critical parameters to dstfile
+
+  cinfo_out.err = jpeg_std_error(&jerr_out);
+  jpeg_create_compress(&cinfo_out);
+  jpeg_stdio_dest(&cinfo_out, fp_out);
+
+  if(srcfile != NULL) // copy critical parameters to dstfile
     jpeg_copy_critical_parameters((j_decompress_ptr)&cinfo_in,&cinfo_out);
+
+  cinfo_out.image_width = image_dims[0];
+  cinfo_out.image_height = image_dims[1];
+  cinfo_out.in_color_space = in_color_space;
+  if(in_components >= 0) cinfo_out.input_components = in_components;
+  cinfo_out.num_components = cinfo_out.input_components;
+
+  if(srcfile == NULL) // set defaults
+    jpeg_set_defaults(&cinfo_out);
+
+  // set sampling factors
+  int chroma_factor[2];
+  if(samp_factor != NULL) {
+    chroma_factor[0] = *(samp_factor + 0);
+    chroma_factor[1] = *(samp_factor + 1);
+    for(int comp = 0; comp < cinfo_out.num_components; comp++) {
+      cinfo_out.comp_info[comp].h_samp_factor = *(samp_factor + comp*2 + 0);
+      cinfo_out.comp_info[comp].v_samp_factor = *(samp_factor + comp*2 + 1);
+    }
+  } else {
+    chroma_factor[0] = cinfo_out.comp_info[0].h_samp_factor;
+    chroma_factor[1] = cinfo_out.comp_info[0].v_samp_factor;
   }
+  // for(int comp = 0; comp < cinfo.num_components; comp++) {
+  //     *(samp_factor + comp*2 + 0) = cinfo.comp_info[comp].h_samp_factor;
+  //     *(samp_factor + comp*2 + 1) = cinfo.comp_info[comp].v_samp_factor;
+  //   }
+  // if(samp_factor != NULL) {
+  //   int J_factor = *(samp_factor + 0);
+  //   int a_factor = *(samp_factor + 1);
+  //   int b_factor = *(samp_factor + 2);
+    
+  //   cinfo_out.comp_info[0].h_samp_factor = chroma_factor[0] = J_factor / a_factor;
+  //   cinfo_out.comp_info[0].v_samp_factor = chroma_factor[1] = (int)(a_factor == b_factor) + 1;
+  //   cinfo_out.comp_info[1].h_samp_factor = cinfo_out.comp_info[1].v_samp_factor = 1;
+  //   cinfo_out.comp_info[2].h_samp_factor = cinfo_out.comp_info[2].v_samp_factor = 1;
+  // } else {
+  //   chroma_factor[0] = cinfo_out.comp_info[0].h_samp_factor;
+  //   chroma_factor[1] = cinfo_out.comp_info[0].v_samp_factor;
+  // }
+  //fprintf(stderr, "chroma factors %d %d\n", chroma_factor[0], chroma_factor[1]);
+  //for(int ch = 0; ch < 3; ch++) {
+  //  fprintf(stderr, "sampling factors(%d) %d %d\n", ch, cinfo_out.comp_info[ch].v_samp_factor, cinfo_out.comp_info[ch].h_samp_factor);
+  //}
+  
+    // for(int comp = 0; comp < cinfo_out.input_components; comp++) {
+    //   //cinfo_out.comp_info[comp].h_samp_factor
+    //   //int J_factor,a_factor,b_factor;
+    //   //J,a,b = samp_factor
+    //   //  #samp_factor = np.array([
+    //   //  #    [int(J / a), int(a == b) + 1],
+    //   //fprintf(stderr, "scale %d: %d %d\n", comp, cinfo_out.comp_info[comp].h_samp_factor, cinfo_out.comp_info[comp].v_samp_factor);
+    //   cinfo_out.comp_info[comp].h_samp_factor = *(samp_factor + comp*2 + 0);
+    //   cinfo_out.comp_info[comp].v_samp_factor = *(samp_factor + comp*2 + 1);
+    //   //fprintf(stderr, "scale %d': %d %d\n", comp, cinfo_out.comp_info[comp].h_samp_factor, cinfo_out.comp_info[comp].v_samp_factor);
+    // }
+  
 
   // write qt
-  if(qt != NULL) {
-    for(int ch = 0; ch < 2; ch++) {
-      memcpy((void *)cinfo_out.quant_tbl_ptrs[ch]->quantval, (void *)(qt + ch*64), sizeof(short)*64);
-    }
+  if(qt != NULL)
+    for(int ch = 0; ch < 2; ch++)
+      for(int i = 0; i < 64; i++)
+        cinfo_out.quant_tbl_ptrs[ch]->quantval[i]/*(i&7)*8+(i>>3)*/ = qt[ch*64 + i];
+        //qt[ch*64 + i] = cinfo_out.quant_tbl_ptrs[ch]->quantval[(i&7)*8+(i>>3)];
+      //memcpy((void *)cinfo_out.quant_tbl_ptrs[ch]->quantval, (void *)(qt + ch*64), sizeof(short)*64);
   // write quality
-  } else if(quality > 0) {
+  else if(quality > 0)
     jpeg_set_quality(&cinfo_out, quality, TRUE);
-  }
 
-  // dct
-  if(dct != NULL) {
+  //fprintf(stderr, "After setting of qt.\n");
 
-    // allocate coefficient buffers
-    jvirt_barray_ptr *coeffs_array = (jvirt_barray_ptr *)(cinfo_out.mem->alloc_small)(
+  // DCT coefficients
+  jvirt_barray_ptr *coeffs_array;
+  if(srcfile != NULL) { // copy from source
+    //fprintf(stderr, "Copy from source.\n");
+    coeffs_array = jpeg_read_coefficients(&cinfo_in);
+  } else { // allocate new
+    //fprintf(stderr, "Create new array.\n");
+    coeffs_array = (jvirt_barray_ptr *)(cinfo_out.mem->alloc_small)(
       (j_common_ptr)&cinfo_out,
       JPOOL_IMAGE,
       sizeof(jvirt_barray_ptr) * cinfo_out.num_components
     );
-    for(int ch = 0; ch < 3; ch++) {
+    //fprintf(stderr, "Set the coefficients. Image: %dx%d\n", cinfo_out.image_width, cinfo_out.image_height);
+    for(int ch = 0; ch < (cinfo_out.num_components); ch++) { // channel iterator
       jpeg_component_info* comp_ptr = cinfo_out.comp_info + ch;
-      long v_samp_factor = *(samp_factor + ch*2 + 0);
-      long h_samp_factor = *(samp_factor + ch*2 + 1);
-      comp_ptr->height_in_blocks = cinfo_out.image_width / 8 * h_samp_factor;// / 4;
-      comp_ptr->width_in_blocks = cinfo_out.image_height / 8 * v_samp_factor;// / 4;
+      //long v_samp_factor = *(samp_factor + ch*2 + 0);
+      //long h_samp_factor = *(samp_factor + ch*2 + 1);
+      comp_ptr->width_in_blocks = (JDIMENSION)ceil(((double)cinfo_out.image_width) / 8);// * comp_ptr->h_samp_factor / 4;
+      comp_ptr->height_in_blocks = (JDIMENSION)ceil(((double)cinfo_out.image_height) / 8);// * comp_ptr->v_samp_factor / 4;
+      if(ch > 0) {
+        comp_ptr->width_in_blocks = ceil(((double)comp_ptr->width_in_blocks) / chroma_factor[0]);
+        comp_ptr->height_in_blocks = ceil(((double)comp_ptr->height_in_blocks) / chroma_factor[1]);
+      }
       coeffs_array[ch] = (cinfo_out.mem->request_virt_barray)(
         (j_common_ptr)&cinfo_out,
         JPOOL_IMAGE,
         TRUE,
         (JDIMENSION)jround_up(comp_ptr->width_in_blocks,  //component size in dct blocks (ignoring mcu)
-		                          h_samp_factor),   //round up is important, if border MCUs are not completely needed
-		    (JDIMENSION)jround_up(comp_ptr->height_in_blocks,
-		                          v_samp_factor),
-		    (JDIMENSION)v_samp_factor
+		                          comp_ptr->h_samp_factor),   //round up is important, if border MCUs are not completely needed
+        (JDIMENSION)jround_up(comp_ptr->height_in_blocks,
+		                          comp_ptr->v_samp_factor),
+		    (JDIMENSION)comp_ptr->v_samp_factor
       );
-    }
-    jpeg_write_coefficients(&cinfo_out,coeffs_array);
 
-    // write dct
+    }
+    //fprintf(stderr, "ready to write coefficient\n");
+  }
+  #if JPEG_LIB_VERSION >= 80
+  jpeg_calc_jpeg_dimensions(&cinfo_out);
+  #endif
+  //fprintf(stderr, "before writing coefficients\n");
+  jpeg_write_coefficients(&cinfo_out,coeffs_array);
+  //fprintf(stderr, "written coefficients\n");
+  // write DCT coefficients
+  if(dct != NULL) {
     JBLOCKARRAY buffer_one;
     JCOEFPTR blockptr_one;
     int HblocksY = cinfo_out.comp_info->height_in_blocks; // max height
@@ -282,20 +365,17 @@ int write_jpeg_dct(
       int Hblocks = comp_ptr->height_in_blocks; // max height
       int Wblocks = comp_ptr->width_in_blocks; // max width
       for(int h = 0; h < Hblocks; h++) { // height iterator
-        buffer_one = (cinfo_out.mem->access_virt_barray)((j_common_ptr)&cinfo_out, coeffs_array[ch], h, (JDIMENSION)1, FALSE);
+        //fprintf(stderr, "accessing ch%d h%d/[H%d,W%d]\n", ch, h, Hblocks, Wblocks);
+        buffer_one = (cinfo_out.mem->access_virt_barray)((j_common_ptr)&cinfo_out, coeffs_array[ch], h, (JDIMENSION)1, TRUE);
         for(int w = 0; w < Wblocks; w++) { // width iterator
           blockptr_one = buffer_one[0][w];
-          for(int bh = 0; bh < 8; bh++) {
-            for(int bw = 0; bw < 8; bw++) {
-              int i = bw*8 + bh;
-              blockptr_one[bh*8 + bw] = ((short *)_dct_offset(dct, ch, w, h, WblocksY, HblocksY))[i];
-            }
-          }
+          for(int bh = 0; bh < 8; bh++)
+            for(int bw = 0; bw < 8; bw++)
+              blockptr_one[bh*8 + bw] = ((short *)_dct_offset(dct, ch, w, h, WblocksY, HblocksY))[bw*8 + bh];
         }
       }
     }
   }
-  
   // cleanup
   jpeg_finish_compress( &cinfo_out );
   jpeg_destroy_compress( &cinfo_out );
@@ -313,6 +393,8 @@ int write_jpeg_dct(
 int read_jpeg_spatial(
   const char *srcfile,
   unsigned char *rgb,
+  unsigned char *colormap, // colormap used
+  unsigned char *in_colormap, // colormap to use
   int out_color_space,
   int dither_mode,
   int dct_method,
@@ -330,6 +412,7 @@ int read_jpeg_spatial(
   // set parameters
   if(out_color_space >= 0) cinfo.out_color_space = out_color_space;
   else cinfo.out_color_space = cinfo.jpeg_color_space;
+  //fprintf(stderr, "cjpeglib.c: out_color_space: %d %d\n", out_color_space, cinfo.out_color_space);
   
   if(dither_mode >= 0) cinfo.dither_mode = dither_mode;
   if(dct_method >= 0) cinfo.dct_method = dct_method;
@@ -337,6 +420,20 @@ int read_jpeg_spatial(
   cinfo.do_fancy_upsampling   = 0 != (flags & DO_FANCY_UPSAMPLING);
   cinfo.do_block_smoothing    = 0 != (flags & DO_BLOCK_SMOOTHING);
   cinfo.quantize_colors       = 0 != (flags & QUANTIZE_COLORS);
+  
+  unsigned char *cmap[256];
+  if(in_colormap != NULL)
+    for(int i = 0; i < 256; i++) {
+      cmap[i] = in_colormap + i*3;
+      //if(i < 3) fprintf(stderr, " %d %d |", i, cmap[i][0]);
+      //if(i == 255) fprintf(stderr, "\n");
+    }
+  if(flags & QUANTIZE_COLORS) {
+    cinfo.actual_number_of_colors = 256; // TODO: parametrized
+    cinfo.desired_number_of_colors = 256;
+    if(in_colormap != NULL) cinfo.colormap = (unsigned char**)cmap;
+  }
+  
   cinfo.progressive_mode      = 0 != (flags & PROGRESSIVE_MODE);
   cinfo.arith_code            = 0 != (flags & ARITH_CODE);
   cinfo.CCIR601_sampling      = 0 != (flags & CCIR601_SAMPLING);
@@ -344,15 +441,25 @@ int read_jpeg_spatial(
   cinfo.enable_1pass_quant    = 0 != (flags & ENABLE_1PASS_QUANT);
   cinfo.enable_external_quant = 0 != (flags & ENABLE_EXTERNAL_QUANT);
   cinfo.enable_2pass_quant    = 0 != (flags & ENABLE_2PASS_QUANT);
-
+  
   // decompress
   (void)jpeg_start_decompress(&cinfo);
-
-  // read
+  // read pixels
   unsigned char *rowptr = rgb;
+  unsigned short stride = (flags & QUANTIZE_COLORS)?1:cinfo.out_color_components;
   while(cinfo.output_scanline < cinfo.output_height) {
     jpeg_read_scanlines(&cinfo, &rowptr, 1);
-    rowptr += cinfo.output_width * cinfo.out_color_components;
+    rowptr += cinfo.output_width * stride;
+  }
+  // read quantization colormap
+  if(flags & QUANTIZE_COLORS) {
+    int N = cinfo.out_color_components;
+    for(int ch=0; ch < N; ch++) {
+      for(int i=0; i < 256; i++) {
+        colormap[ch*256 + i] = cinfo.colormap[ch][i];
+        //colormap[ch*256 + i] = cinfo.colormap[i][ch];
+      }
+    }
   }
 
   // cleanup
@@ -393,13 +500,33 @@ int write_jpeg_spatial(
   jpeg_stdio_dest(&cinfo, fp);
 
   // set basic parameters
-  cinfo.image_width = image_dims[0];
-  cinfo.image_height = image_dims[1];
-  cinfo.in_color_space = in_color_space;
-  if(in_components >= 0) cinfo.input_components = in_components;
-  cinfo.num_components = cinfo.input_components;
-  jpeg_set_defaults(&cinfo);
-  
+  if(srcfile == NULL) {
+    cinfo.image_width = image_dims[0];
+    cinfo.image_height = image_dims[1];
+    if(in_color_space >= 0)
+      cinfo.in_color_space = in_color_space;
+    //fprintf(stderr, "writing 1 from cs %d to %d\n", cinfo.in_color_space, cinfo.jpeg_color_space);
+    if(in_components >= 0)
+      cinfo.input_components = in_components;
+    cinfo.num_components = cinfo.input_components;
+    jpeg_set_defaults(&cinfo);
+  }
+
+  // set advanced parameters
+  if(dct_method >= 0) cinfo.dct_method = dct_method;
+  int chroma_factor[2];
+  if(samp_factor != NULL) {
+    chroma_factor[0] = *(samp_factor + 0);
+    chroma_factor[1] = *(samp_factor + 1);
+    for(int comp = 0; comp < cinfo.num_components; comp++) {
+      cinfo.comp_info[comp].h_samp_factor = *(samp_factor + comp*2 + 0);
+      cinfo.comp_info[comp].v_samp_factor = *(samp_factor + comp*2 + 1);
+    }
+  } else {
+    chroma_factor[0] = cinfo.comp_info[0].h_samp_factor;
+    chroma_factor[1] = cinfo.comp_info[0].v_samp_factor;
+  }
+
   // copy parameters
   struct jpeg_decompress_struct cinfo_in;
   struct jpeg_error_mgr jerr_in;
@@ -409,35 +536,34 @@ int write_jpeg_spatial(
     if((fp_in = _read_jpeg(srcfile, &cinfo_in, &jerr_in)) == NULL) return 0;
     // decompress
     (void)jpeg_start_decompress(&cinfo_in);
-    
     jpeg_copy_critical_parameters((j_decompress_ptr)&cinfo_in, (j_compress_ptr)&cinfo);
   }
 
-  // set advanced parameters
-  if(dct_method >= 0) cinfo.dct_method = dct_method;
-  if(samp_factor != NULL)
-    for(int comp = 0; comp < cinfo.input_components; comp++) {
-      cinfo.comp_info[comp].h_samp_factor = *(samp_factor + comp*2 + 0);
-      cinfo.comp_info[comp].v_samp_factor = *(samp_factor + comp*2 + 1);
-    }
-
-  
   if(qt != NULL) {
-    for(int ch = 0; ch < 2; ch++) {
-      memcpy((void *)cinfo.quant_tbl_ptrs[ch]->quantval, (void *)(qt + ch*64), sizeof(short)*64);
-    }
-    // unsigned qt_u[64];
-    // for(int i = 0; i < 64; i++) qt_u[i] = qt[i];
-    // jpeg_add_quant_table(&cinfo, 2, qt_u, 100, TRUE);
-    // cinfo.comp_info[0].quant_tbl_no = 2;
-    // for(int i = 0; i < 64; i++) qt_u[i] = qt[64 + i];
-    // jpeg_add_quant_table(&cinfo, 3, qt_u, 100, TRUE);
-    // cinfo.comp_info[1].quant_tbl_no = 3;
-    // cinfo.comp_info[2].quant_tbl_no = 3;
+    unsigned qt_u[64];
+    // component 0
+    for(int i = 0; i < 64; i++) qt_u[i] = qt[i]; // (i&7)*8+(i>>3)
+    jpeg_add_quant_table(&cinfo, 0, qt_u, 100, FALSE);
+    cinfo.comp_info[0].component_id = 0;
+    cinfo.comp_info[0].quant_tbl_no = 0;
+    // component 1
+    for(int i = 0; i < 64; i++) qt_u[i] = qt[64 + i]; // (i&7)*8+(i>>3)
+    jpeg_add_quant_table(&cinfo, 1, qt_u, 100, FALSE);
+    cinfo.comp_info[1].component_id = 1;
+    cinfo.comp_info[1].quant_tbl_no = 1;
+    // component 2
+    cinfo.comp_info[2].component_id = 2;
+    cinfo.comp_info[2].quant_tbl_no = 1;
   } else if(quality >= 0) {
-    jpeg_set_quality(&cinfo, quality, TRUE);
+    jpeg_set_quality(&cinfo, quality, FALSE);
   }
   if(smoothing_factor >= 0) cinfo.smoothing_factor = smoothing_factor;
+  if(in_color_space >= 0) {
+    cinfo.in_color_space = in_color_space;
+    //jpeg_set_colorspace(&cinfo, in_color_space);
+    //fprintf(stderr, "writing 2 from cs %d to %d\n", cinfo.in_color_space, cinfo.jpeg_color_space);
+  }
+  //fprintf(stderr, "colorspace conversion %d -> %d\n", cinfo.in_color_space, cinfo.jpeg_color_space);
   
   cinfo.progressive_mode   = 0 != (flags & PROGRESSIVE_MODE);
   cinfo.optimize_coding    = 0 != (flags & OPTIMIZE_CODING);
@@ -451,7 +577,6 @@ int write_jpeg_spatial(
   // write data
   unsigned char *rowptr = rgb;
   jpeg_start_compress(&cinfo, TRUE);
-
   for(unsigned h = 0; h < cinfo.image_height; h++) {
     jpeg_write_scanlines(&cinfo, &rowptr, 1);
     rowptr += cinfo.image_width * cinfo.input_components;
@@ -466,7 +591,6 @@ int write_jpeg_spatial(
     fclose( fp_in );
   }
   
-  
   return 1;
 }
 
