@@ -4,6 +4,7 @@ Module containing conversion of C-structures into numpy and back and memory mana
 
 import ctypes
 import logging
+import math
 import numpy as np
 import tempfile
 import warnings
@@ -13,7 +14,7 @@ from ._timer import Timer
 
 class JPEG:
     cjpeglib = CJpegLib
-    def __init__(self, srcfile):
+    def __init__(self, srcfile=None):
         """Constructs the :class:`JPEG` object.
         
         :param srcfile: Path to a source file in JPEG format.
@@ -24,31 +25,36 @@ class JPEG:
 
         >>> im = jpeglib.JPEG("input.jpeg")
         """
+        t = Timer("allocate default")
         # set filename
         self.srcfile = srcfile
         # allocate resources
         self._dct_dims = (ctypes.c_int*6)()
         self._dims = (ctypes.c_int*3)()
         self._num_components = (ctypes.c_int*1)()
-        self._color_space = (ctypes.c_int*1)()
-        self._samp_factor = ((ctypes.c_int*2)*3)()
+        self._samp_factor = self._parse_samp_factor()
+        self.dct_channels = 3
+        self.color_space = 'JCS_RGB'
+        del t
+        #self.color_space = [k for k,v in self.J_COLOR_SPACE.items() if v[0] == self._color_space[0]][0]
         # get image info
-        self._read_info()   
-        # allocate
-        self._im_spatial = self._allocate_spatial()
-        self._im_dct = self._allocate_dct()
-        self._im_qt = ((ctypes.c_short*64)*(self.dct_channels-1))()
+        if srcfile is not None:
+            self._read_info()
+            self._allocate() # allocate
 
-    def read_dct(self):
-        """Reads the non-quantized DCT coefficients and quantization tables of the source file.
+    def read_dct(self, quantized=False):
+        """Reads the DCT coefficients and quantization tables of the source file.
 
-        In the return values, Y is non-quantized DCT luminance tensor
-        of shape (1,W/8,H/8,8,8), CbCr is non-quantized DCT chrominance tensor
-        of shape (2,W/8,H/8,8,8) and qt is a tensor with quantization tables of shape (2,8,8).
+        In the return values, Y is DCT luminance tensor of shape (1,W/8,H/8,8,8),
+        CbCr is DCT chrominance tensor of shape (2,W/8,H/8,8,8), both not-quantized by default.
+        qt is a tensor with quantization tables of shape (2,8,8).
         Check the examples below.
 
+        :param quantized: Indicates whether the output DCT coefficients are quantized or not, False by default.
+        :type quantized: bool, optional 
         :return: lumo DCT Y, chroma DCT CbCr, quantization table qt
         :rtype: tuple
+        :raises [IOError]: When source file was not given in constructor.
 
         :Example:
 
@@ -67,22 +73,20 @@ class JPEG:
         >>> qt[1] # Cb
         >>> qt[1] # Cr
         """
-        #t = Timer('reading %s DCT', self.srcfile) # log execution time
         # execute
         Y,CbCr,qt = self._read_dct(self.srcfile)
+        # quantization
+        if quantized: Y,CbCr = Y*qt[0],CbCr*qt[1]
         # result
         return Y,CbCr,qt
 
-    def write_dct(self, dstfile, Y=None, CbCr=None, qt=None, in_color_space=None, samp_factor=None):
+    def write_dct(self, dstfile, Y=None, CbCr=None, qt=None, quantized=False, in_color_space=None, samp_factor=None):
         """Writes DCT coefficients to a file.
         
         The shape and arrangement of the parameters Y and CbCr is the same
         as what is returned from :meth:`jpeg.JPEG.read_dct`.
-        Y is non-quantized DCT luminance tensor of shape (1,W/8,H/8,8,8),
-        CbCr is non-quantized DCT chrominance tensor of shape (2,W/8,H/8,8,8)
-        and qt is either a tensor with quantization tables of shape (2,8,8)
-        or a quality scalar integer.
-
+        Y is DCT luminance tensor of shape (1,W/8,H/8,8,8), CbCr is DCT chrominance tensor of shape (2,W/8,H/8,8,8),
+        both unquantized by default. qt is either a tensor with quantization tables of shape (2,8,8) or a quality scalar integer.
 
         :param dstfile: Destination file.
         :type dstfile: str
@@ -92,6 +96,8 @@ class JPEG:
         :type CbCr: numpy.ndarray, optional
         :param qt: Quantization tables or quality.
         :type qt: numpy.ndarray | int, optional
+        :param quantized: Indicates whether the input DCT coefficients are quantized or unquantized, False by default.
+        :type quantized: bool, optional 
         :param in_color_space: Input color space. Must be key of :class:`jpeg.JPEG.J_COLOR_SPACE`. According to source by default.
         :type in_color_space: str, optional
         :param samp_factor: Sampling factor. None, tuple of three ints or tuples of two ints. According to source by default.
@@ -112,9 +118,8 @@ class JPEG:
         If neither quality nor quantization tables are specified,
         library uses the quantization table from the source.
         """
-        #t = Timer('writing %s DCT', dstfile) # log execution time
         # execute
-        self._write_dct(dstfile, Y, CbCr, qt, samp_factor)
+        self._write_dct(dstfile, Y, CbCr, qt, quantized, in_color_space, samp_factor)
         self._im_dct = None # free DCT buffer
         self._im_spatial = None # free DCT buffer
 
@@ -141,7 +146,7 @@ class JPEG:
         "JDCT_FLOAT": 2, # floating-point method
     }
 
-    def read_spatial(self, out_color_space=None, dither_mode=None, dct_method=None, flags=[]):
+    def read_spatial(self, out_color_space=None, dither_mode=None, dct_method=None, colormap=None, flags=[]):
         """Decompresses the file into the spatial domain. 
         
         :param out_color_space: Output color space. Must be key of J_COLOR_SPACE.
@@ -150,6 +155,8 @@ class JPEG:
         :type dither_mode: str, optional
         :param dct_method: DCT method. Must be key of :class:`jpeg.JPEG.J_DCT_METHOD`. Using default from libjpeg by default.
         :type dct_method: str, optional
+        :param colormap: Colormap to use for color quantization.
+        :type colormap: np.array, optional
         :param flags: Bool decompression parameters as str to set to be true. Using default from libjpeg by default.
         :type flags: list, optional
         :return: Spatial representation of the source image.
@@ -162,7 +169,7 @@ class JPEG:
         """
         #t = Timer('reading %s spatial', self.srcfile) # log execution time
         # execute
-        spatial = self._read_spatial(self.srcfile, out_color_space, dither_mode, dct_method, flags)
+        spatial = self._read_spatial(self.srcfile, out_color_space, dither_mode, dct_method, colormap, flags)
         self._im_dct = None # free DCT buffer
         # result
         return spatial
@@ -202,16 +209,24 @@ class JPEG:
         self._im_dct = None # free DCT buffer
         self._im_spatial = None # free spatial buffer
 
-    def to_spatial(self, Y=None, CbCr=None, **kw): #, qt=None):
+    def source_color_space(self):
+        """Returns the color space of the source file."""
+        return self.color_space
+
+    def to_spatial(self, Y=None, CbCr=None, qt=None, quantized=False, **kw): #, qt=None):
         """Converts DCT representation to spatial.
         
         Performs decompression from DCT coefficients to spatial representation.
         Uses temporary file to save and read from as libjpeg does not have a direct handler.
         
-        :param Y: Luminance DCT.
-        :type Y: numpy.ndarray
+        :param Y: Lumo DCT.
+        :type Y: numpy.ndarray, optional
         :param CbCr: Chrominance DCT.
-        :type CbCr: numpy.ndarray
+        :type CbCr: numpy.ndarray, optional
+        :param qt: Quantization tables or quality.
+        :type qt: numpy.ndarray | int, optional
+        :param quantized: Indicates whether the input DCT coefficients are quantized or unquantized, False by default.
+        :type quantized: bool, optional
         :return: Spatial representation of DCT coefficients
         :rtype: numpy.ndarray
 
@@ -221,38 +236,81 @@ class JPEG:
         >>> Y,CbCr,qt = im.read_dct()
         >>> spatial = im.to_spatial(Y, CbCr, out_color_space="JCS_RGB")
         """
-        #t = Timer('DCT-RGB conversion')
         if (Y is None or CbCr is None) and self._im_dct is None:
             raise RuntimeError("Call read_dct() before calling to_spatial() or specify Y and CbCr.")
         with tempfile.NamedTemporaryFile() as tmp:
-            self.write_dct(tmp.name, Y=Y, CbCr=CbCr)
-            data = self._read_spatial(tmp.name, **kw)
+            self.write_dct(tmp.name, Y=Y, CbCr=CbCr, qt=qt, quantized=quantized)
+            with JPEG(tmp.name) as im:
+                data = im.read_spatial(**kw)
         self._im_dct = None
         return data
 
     def _read_info(self):
         # get information
+        t = Timer("read_jpeg_info()")
+        _color_space = (ctypes.c_int*1)()
         self.cjpeglib.read_jpeg_info(
             srcfile = self.srcfile,
             dct_dims = self._dct_dims,
             image_dims = self._dims,
             num_components = self._num_components,
             samp_factor = self._samp_factor,
-            jpeg_color_space = self._color_space
+            jpeg_color_space = _color_space
         )
+        del t
         # parse
-        self.dct_channels = 3
+        t = Timer("parse arguments")
         self.channels = self._num_components[0]
         self.dct_shape = np.array([self._dct_dims[i] for i in range(6)], int)\
             .reshape(self.dct_channels, 2)
         self.shape = np.array([self._dims[0], self._dims[1]])
-        self.color_space = [k for k,v in self.J_COLOR_SPACE.items() if v[0] == self._color_space[0]][0]
+        self.color_space = [k for k,v in self.J_COLOR_SPACE.items() if v[0] == _color_space[0]][0]
+        del t
+        #print("self.color_space set to", self.color_space, _color_space[0])
+        #print("init color space", _color_space[0], self.color_space)
 
+    def _allocate(self):
+        t = Timer("allocate()")
+        self._im_spatial,self._im_colormap = self._allocate_spatial()
+        self._im_dct = self._allocate_dct()
+        self._im_qt = ((ctypes.c_short*64)*(self.dct_channels-1))()
+        del t
+
+    def _initialize_info(self, data=None, Y=None, CbCr=None):
+        if data is not None:
+            self.channels = data.shape[2]
+            self.shape = np.array(data.shape[:2])
+        else:
+            self.channels = 3
+            self.shape = np.array([0,0])
+        if Y is not None and CbCr is not None:
+            self._dct_dims[0] = Y.shape[1]
+            self._dct_dims[1] = Y.shape[2]
+            self._dct_dims[2] = CbCr.shape[1]
+            self._dct_dims[3] = CbCr.shape[2]
+            self._dct_dims[4] = CbCr.shape[1]
+            self._dct_dims[5] = CbCr.shape[2]
+            self.shape = np.array([self._dct_dims[1], self._dct_dims[0]]) * 8
+            self.channels = 3
+        else:
+            for i in range(6):
+                self._dct_dims[i] = math.ceil((data.shape[i % 2] / 8))
+                #if self._samp_factor[i // 2][i % 2] != 0:
+                #    self._dct_dims[i] = math.ceil(self._dct_dims[i])# / self._samp_factor[i // 2][i % 2])
+
+        self.dct_shape = np.array([self._dct_dims[i] for i in range(6)], int)\
+            .reshape(self.dct_channels, 2)
+        self._dims = (ctypes.c_int * 3)(self.shape[1], self.shape[0], self.channels)
+
+        self._allocate()
     
     def _read_dct(self, srcfile):
         # allocate
         if self._im_dct is None:
             self._im_dct = self._allocate_dct()
+        # check source file
+        if srcfile is None:
+            raise IOError("source file not given")
         # reading
         self.cjpeglib.read_jpeg_dct(srcfile, self._im_dct, self._im_qt)
         # align qt
@@ -261,15 +319,18 @@ class JPEG:
         # align lumo
         Y = np.ctypeslib.as_array(self._im_dct[:1])
         Y = Y.reshape((*Y.shape[:-1],8,8))
-        # align chroma 
+        # align chroma
         CbCr = np.ctypeslib.as_array(self._im_dct[1:])
         CbCr = CbCr[:,:self.dct_shape[1][0],:self.dct_shape[1][1]]
         CbCr = CbCr.reshape((*CbCr.shape[:-1],8,8))
         # finish
         return Y,CbCr,qt
 
-    def _write_dct(self, dstfile, Y=None, CbCr=None, quality=None, in_color_space=None, samp_factor=None):
-        # TODO: remove copying from source file
+    def _write_dct(self, dstfile, Y=None, CbCr=None, quality=None, quantized=False, in_color_space=None, samp_factor=None):
+        # initialize default
+        samp_factor = self._parse_samp_factor(samp_factor) # sampling factor
+        if self.srcfile is None:
+            self._initialize_info(Y=Y, CbCr=CbCr)
         # allocate
         if self._im_dct is None:
             self._im_dct = self._allocate_dct()
@@ -277,22 +338,27 @@ class JPEG:
         if in_color_space is None:
             in_color_space = self.color_space
         in_color_space,channels = self.J_COLOR_SPACE[in_color_space]
-
+        # quantization
+        if quantized and quality is None:
+            qt = np.ctypeslib.as_array(self._im_qt)
         # align lumo
         if Y is not None:
             Y = Y.reshape((*Y.shape[:-2],64))
-            self._im_dct[:1] = np.ctypeslib.as_ctypes(Y)
+            if quantized:
+                Y /= qt[0]
+            self._im_dct[0] = np.ctypeslib.as_ctypes(Y[0])
         # align chroma
         if CbCr is not None:
             CbCr = CbCr.reshape((*CbCr.shape[:-2],64))
+            if quantized:
+                CbCr /= qt[1]
             _CbCr = np.zeros((2, self.dct_shape[0][0], self.dct_shape[0][1], 64), np.short)
-            _CbCr[:,:int(self.dct_shape[1][0]),:self.dct_shape[1][1]] = CbCr
+            _CbCr[:,:int(self.dct_shape[1][0]),:int(self.dct_shape[1][1])] = CbCr
             self._im_dct[1:] = np.ctypeslib.as_ctypes(_CbCr)
         # quality
         qt,quality,srcfile = self._parse_quality(quality)
-        # sampling factor
-        samp_factor = self._parse_samp_factor(samp_factor)
         # write
+        #print("write_jpeg_dct:", self._dims[:], in_color_space, self.channels, [samp_factor[i] for i in range(3)])
         self.cjpeglib.write_jpeg_dct(
             srcfile        = srcfile,
             dstfile        = dstfile,
@@ -302,27 +368,40 @@ class JPEG:
             in_components  = self.channels,
             samp_factor    = samp_factor,
             qt             = qt,
-            quality        = quality)
+            quality        = quality
+        )
 
 
-    def _read_spatial(self, srcfile, out_color_space, dither_mode, dct_method, flags):
+    def _read_spatial(self, srcfile, out_color_space, dither_mode, dct_method, colormap, flags):
         # parameters
-        if out_color_space is None:
+        out_color_space = out_color_space if out_color_space is not None else None
+        if out_color_space is None and self.color_space is not None:
             out_color_space = self.color_space
-        self.color_space = out_color_space
         color_space,channels = self.J_COLOR_SPACE[out_color_space]
+        if srcfile is None:
+            raise IOError("source file not given")
+        #print("reading with color space:", out_color_space, color_space, 'default', self.color_space)
         dither_mode = self.J_DITHER_MODE[dither_mode]
         dct_method = self.J_DCT_METHOD[dct_method]
+        in_cmap = None
+        if colormap is not None:
+            if 'QUANTIZE_COLORS' not in flags:
+                flags.append('QUANTIZE_COLORS')
+            in_cmap = np.ascontiguousarray(colormap)
+            in_cmap = np.ctypeslib.as_ctypes(in_cmap.astype(np.ubyte))
         # allocate
         if self._im_spatial is None or channels != self.channels:
-            self.channels = channels
-            self._im_spatial = self._allocate_spatial()
+            if channels > 0:
+                self.channels = channels
+            self._im_spatial,self._im_colormap = self._allocate_spatial()
         else: self.channels = channels
-        
+
         # call
         self.cjpeglib.read_jpeg_spatial(
             srcfile         = srcfile,
             rgb             = self._im_spatial,
+            colormap        = self._im_colormap,
+            in_colormap     = in_cmap,
             out_color_space = color_space,
             dither_mode     = dither_mode,
             dct_method      = dct_method,
@@ -330,18 +409,33 @@ class JPEG:
         )
         # align rgb
         data = np.ctypeslib.as_array(self._im_spatial).astype(np.ubyte)
-        data = data.reshape(data.shape[2],-1,self.channels)
+        if 'QUANTIZE_COLORS' in flags:
+            data = data.reshape(self.channels,data.shape[2],-1)[0]
+            # parse colormap
+            colormap = np.ctypeslib.as_array(self._im_colormap)
+            colormap = colormap.reshape(colormap.shape[1], self.channels)
+            # index to color
+            data = np.array([[colormap[i] for i in row] for row in np.array(data)])
+        else:
+            #print(data.shape, self.channels)
+            data = data.reshape(data.shape[2],-1,self.channels)
+        # quantization
+        #if 'QUANTIZE_COLORS' in flags:
+        #    data = data[:,:,0]
         # finish
         return data
     
     def _write_spatial(self, dstfile, data, in_color_space, dct_method, samp_factor, quality, smoothing_factor, flags):
         """"""
-        # parameters
-        if in_color_space is None:
-            in_color_space = self.color_space
-        in_color_space,channels = self.J_COLOR_SPACE[in_color_space]
-        dct_method = self.J_DCT_METHOD[dct_method]
+        # initialize default
         self._samp_factor = self._parse_samp_factor(samp_factor)
+        if self.srcfile is None:
+            self._initialize_info(data=data)
+        # parameters
+        in_color_space = in_color_space if in_color_space is not None else self.color_space
+        color_space,channels = self.J_COLOR_SPACE[in_color_space]
+        #print("writing with color space:", in_color_space, color_space, channels)
+        dct_method = self.J_DCT_METHOD[dct_method]
         qt,quality,srcfile = self._parse_quality(quality)
         # spatial buffer
         if data is not None:
@@ -351,13 +445,12 @@ class JPEG:
         elif self._im_spatial is None:
             warnings.warn('Writing unsuccessful, call read_spatial() before calling write_spatial() or specify data parameter.', RuntimeWarning)
             return
-
         self.cjpeglib.write_jpeg_spatial(
             srcfile          = self.srcfile,
             dstfile          = dstfile,
             rgb              = self._im_spatial,
             image_dims       = self._dims,
-            in_color_space   = in_color_space,
+            in_color_space   = color_space,
             in_components    = self.channels,
             dct_method       = dct_method,
             samp_factor      = self._samp_factor,
@@ -396,26 +489,32 @@ class JPEG:
                                   * self.dct_shape[0][0])
                                   * self.dct_channels)()
     def _allocate_spatial(self):
-        return (((ctypes.c_ubyte * self.shape[1])
+        return ((((ctypes.c_ubyte * self.shape[1])
                                  * self.shape[0])
-                                 * self.channels)()
-    def _parse_samp_factor(self, samp_factor):
-        if samp_factor is not None:
-            samp_factor = list(samp_factor)
-            for i,f in enumerate(samp_factor):
-                if isinstance(f,int):
-                    self._samp_factor[i][0] = f
-                    self._samp_factor[i][1] = f
-                else:
-                    self._samp_factor[i][0] = f[0]
-                    self._samp_factor[i][1] = f[1]
+                                 * self.channels)(),
+                ((ctypes.c_ubyte * 256) * self.channels)())
+    def _parse_samp_factor(self, samp_factor=None):
+        if samp_factor is None:
+            samp_factor = [[2,2],[1,1],[1,1]]#[4, 2, 2]
+
+        samp_factor = np.array(samp_factor, dtype=np.int32)
+        
+        #samp_factor = list(samp_factor)
+        #J,a,b = samp_factor
+        #samp_factor = np.array([
+        #    [int(J / a), int(a == b) + 1],
+        #    [1, 1],
+        #    [1, 1]
+        #], dtype=np.int32)
+
+        self._samp_factor = np.ctypeslib.as_ctypes(samp_factor)
         return self._samp_factor
     def _parse_quality(self, quality):
         if quality is None: # not specified
-            qt = None
+            qt = self._im_qt
             srcfile = self.srcfile
         else:
-            srcfile = None
+            srcfile = self.srcfile #None
             # numeric quality
             try:
                 quality = int(quality)
@@ -428,22 +527,6 @@ class JPEG:
         return qt,quality,srcfile
 
 
-        if quality is not None:
-            srcfile = None
-            # numeric quality
-            try:
-                quality = int(quality)
-                qt = None
-            # quantization table
-            except:
-                qt = np.array(quality, dtype=np.uint16).reshape(2,64)
-                qt = self._im_qt = np.ctypeslib.as_ctypes(qt)
-                quality = -1
-        else: # not specified
-            srcfile = self.srcfile
-            qt = None
-
-
 
 
 __all__ = ["JPEG"]
\ No newline at end of file
